% =================================================================
:- pred constr(bool).
:- mode constr(in).
:- ignore constr/1.
:- type bintree_int ---> leaf ; node(int,bintree_int,bintree_int).
% This is a binary tree supporting a heap. 
% leaf is the EMPTY bintree_int. Nodes may have DUPLICATE integer values.
% ================================================================
% Program: descending sorting.
:- pred heapsort(list(int),list(int)).
:- mode heapsort(in,out).

heapsort(L,SL) :-
  list_to_heap(L,H), heap_to_list(H,SL).
% ===============================================================
% making a heap out of a list.
% Program predicate: not be a catamorphism.
:- pred list_to_heap(list(int),bintree_int).
:- mode list_to_heap(in,out).

list_to_heap([], leaf).
list_to_heap([X|Xs], Heap) :-
  list_to_heap(Xs, HeapXs), 
  insert_heap(X, HeapXs, Heap).

% -------
:- pred insert_heap(int,bintree_int,bintree_int).
:- mode insert_heap(in,in,out).

insert_heap(X, leaf, node(X,leaf,leaf)).
insert_heap(X, node(Top,L,R), node(Top,R,L1)) :- % Torsion! node(Top,R,L1), not node(Top,L1,R), and
          constr(X=<Top), insert_heap(X,L,L1).   % top unchanged. Insert X in the left son-node.
insert_heap(X, node(Top,L,R), node(X,R,L1)) :-   % Torsion! node(Top,R,L1), not node(Top,L1,R), and
          constr(X>Top), insert_heap(Top,L,L1).  % X on top. Insert Top in the left son-node.
% ----------------------------------
% -- The NEW insertion is done, in an alternate way,
% -- into the left child and into the right child.
% -- The resulting heap is **balanced** (as it should be for complexity reasons in heapsort), that is,
% -- the depths of any two leaves differ at most by 1 unit.
% In the last level of the heap, the leaves are not necessarily contiguous, i.e.,
% they are not necessarily flushed to the left or to the right of the frontier of the tree.

:- pred mkbtree(int,bintree_int,bintree_int,bintree_int).
:- mode mkbtree(in,in,in,out).

mkbtree(Top,Left,Right, node(Top,Left,Right)).

% ----------------------------------
% Making a weakly-descending list out of a heap.
% Program predicate: need not be a catamorphism.
:- pred heap_to_list(bintree_int,list(int)).
:- mode heap_to_list(in,out).   

heap_to_list(leaf, []).
heap_to_list(node(Top,leaf,leaf), [Top]).                                    
heap_to_list(node(Top,node(LTop,LL,LR),leaf), [Top|Tail]) :-                   
    mkbtree(LTop,LL,LR,NewHeap),
    heap_to_list(NewHeap,Tail).
heap_to_list(node(Top,leaf,node(RTop,RL,RR)), [Top|Tail]) :-                  
    mkbtree(RTop,RL,RR,NewHeap),
    heap_to_list(NewHeap,Tail).
heap_to_list(node(Top,node(LT,LL,LR),node(RT,RL,RR)), [Top|Tail]) :-        
     mkbtree(LT,LL,LR,NewLHeap), mkbtree(RT,RL,RR,NewRHeap),
     heap_merge(NewLHeap,NewRHeap,MergedHeap),
     heap_to_list(MergedHeap,Tail).

:- pred heap_merge(bintree_int,bintree_int,bintree_int).
:- mode heap_merge(in,in,out).                                              % heap_merge is a total function

heap_merge(leaf,leaf, leaf).                           
heap_merge(node(LTop,LL,LR),leaf, node(LTop,LL,LR)).   
heap_merge(leaf,node(RTop,RL,RR), node(RTop,RL,RR)).   
heap_merge(node(LTop,LL,LR),node(RTop,RL,RR), node(RTop,node(LTop,LL,LR),MergedHeap)) :- 
          LTop =< RTop, heap_merge(RL,RR,MergedHeap).  
heap_merge(node(LTop,LL,LR),node(RTop,RL,RR), node(LTop,MergedHeap,node(RTop,RL,RR))) :- 
          LTop > RTop, heap_merge(LL,LR,MergedHeap).  

% ===============================================================
% Catamorphisms.
% size of a binary-tree.
:- pred btsize(bintree_int,int).
:- mode btsize(in,out).
:- cata btsize/2-1.

btsize(leaf,0).
btsize(node(T,L,R),S) :- constr( S=SL+SR+1 ), btsize(L,SL), btsize(R,SR).

% --- length of a list
:- pred length(list(int),int).
:- mode length(in,out).
:- cata length/2-1.

length([],Res) :- constr( Res=0 ).
length([H|T],Res) :-
    constr( Res=ResT+1 ),
    length(T,ResT).

% ===============================================================
% ff1. Contract on heapsort: length         
:- pred ff1.        
ff1 :-
   constr(~(N1=N2)),
   length(L,N1), length(S,N2),
   heapsort(L,S).
% ---------------

% Contract on list_to_heap:               

:- pred ff11.      
ff11 :-
    constr( ~(NL=NH)), length(L,NL), btsize(H,NH),
    list_to_heap(L,H).
    
% ------------------
% Contract on heap_to_list:            

:- pred ff12.                 
ff12 :-
    constr( ~(SH = LL)),
    btsize(H,SH), length(L,LL), 
    heap_to_list(H,L).
    
% ------------------ 
% Contract on insert_heap:               

:- pred ff2.      
ff2 :-
    constr( ~(SH1=SH+1)), 
    btsize(H,SH), btsize(H1,SH1),
    insert_heap(X,H,H1).
% ----------------------------------------------------------------
% Contract on heap_merge:                       
          
:- pred ff3.    
ff3 :-
   constr(~(SHL + SHR = SH1)),
   btsize(HL,SHL), btsize(HR,SHR), btsize(H1,SH1),
   heap_merge(HL,HR,H1).
%----------------------------------------------------------------
% Contract on mkbtree:                  .     
          
:- pred ff4.    
ff4 :-
   constr(~(SHL + SHR + 1 = SH1)),
   btsize(HL,SHL), btsize(HR,SHR), btsize(H1,SH1),
   mkbtree(T,HL,HR,H1).

% ===============================================================
:- query ff1/0.
:- query ff11/0.
:- query ff12/0.
:- query ff2/0.
:- query ff3/0.
:- query ff4/0.
% ===============================================================