% ====================================================
:- pred constr(bool).
:- mode constr(in).
:- ignore constr/1.

% ====================================================
% Program. mergesorting in ascending order.E.g., 3 =< 5 =< 5 =< 8 =< ...
:- pred mergesort(list(int),list(int)).
:- mode mergesort(in,out).
mergesort([], []).
mergesort([H], [H]).
mergesort([H,K|T],S) :-
   cons(K,T,KT),          
   cons(H,KT,HKT),          
   split(HKT,L1,L2),       
   mergesort(L1,S1),       
   mergesort(L2,S2),
   merge(S1,S2,S).         % merging

:- pred cons(int,list(int),list(int)).
:- mode cons(in,in,out).
cons(H,T,[H|T]).

:- pred split(list(int),list(int),list(int)).
:- mode split(in,out,out).
split(L,L1,L2) :- N1=(N div 2),      % integer division is div in SMTLIB.
     length(L,N),                    
     takedrop(N1,L,L1,L2).

:- pred length(list(int),int).
:- mode length(in,out).
length([],L) :- L=0.
length([H|T],L) :-
  L=LT+1,
  length(T,LT).

:- pred takedrop(int,list(int),list(int),list(int)).
:- mode takedrop(in,in,out,out).
takedrop(N,[],[],[]).
takedrop(N,L,[],L) :- N=<0.
takedrop(N,[H|T],[H|T1],L) :-
  N>=1, N1=N-1,         
  takedrop(N1,T,T1,L).  

:- pred merge(list(int),list(int),list(int)).
:- mode merge(in,in,out).
merge([],L,L).                   
merge([X|Xs],[],[X|Xs]).
merge([X|Xs],[Y|Ys],[X|Zs]) :-
     X=<Y, cons(Y,Ys,YYs),
     merge(Xs,YYs,Zs).
merge([X|Xs],[Y|Ys],[Y|Zs]) :-
     X>Y, cons(X,Xs,XXs),
     merge(XXs,Ys,Zs).

% ====================================================
% catamorphisms
:- pred is_asorted(list(int),bool).
:- mode is_asorted(in,out).
:- cata is_asorted/2-1.
is_asorted([],Res) :- constr(Res).
is_asorted([H|T],Res) :-
    hd(T,IsDefT,HdT),
    is_asorted(T,ResT),
    constr(Res=(IsDefT => ((H=<HdT) & ResT))).

%----
:- pred grt_all(int,list(int),bool).
:- mode grt_all(in,in,out).
:- cata grt_all/3-2.
grt_all(N,[],B) :- constr(B).       
grt_all(N,[X|Xs],B) :-              %
	grt_all(N,Xs,B1),                 % N is greater ( > ) of all elements of list [X|Xs]
	constr(B = ((N>X) & B1)).

%----
:- pred leq_all(int,list(int),bool).
:- mode leq_all(in,in,out).
:- cata leq_all/3-2.
leq_all(N,[],B) :- constr(B).       %
leq_all(N,[Y|Ys],B) :-              % N is leq ( =< ) of all elements of list [Y|Ys]
	leq_all(N,Ys,B1),
	constr(B = ((N=<Y) & B1)).

%----
:- pred hd(list(int),bool,int).
:- mode hd(in,out,out).
:- cata hd/3-1.
hd([],IsDef,Hd) :- constr( (~IsDef) & (Hd=0) ).  % Any int value, instead of 0, is ok.
hd([H|T],IsDef,Hd) :- constr( IsDef & (Hd=H) ).


% ---
:- pred listmin(list(int),bool,int).    
:- mode listmin(in,out,out).
:- cata listmin/3-1.
listmin([],IsDef,A) :- constr( ((~IsDef) & (A=0)) ).  % A can be any int.
listmin([H|T],IsDef,Min) :-
         constr( IsDef & (Min = ite(IsDefT, ite(H<MinT,H,MinT), H)) ),
         listmin(T,IsDefT,MinT).

% ====================================================
% verification. 
% Property:
:- pred ff1.
ff1 :-                              % mergesort
   constr( ~(B)), 
   is_asorted(S,B),
   mergesort(L,S).

:- pred ff2.
ff2 :-                              % split
   constr( ~((SmallYs & BigZs) => (BXs = (BYs & BZs))) ),
   is_asorted(Xs,BXs), is_asorted(Ys,BYs), is_asorted(Zs,BZs),
   grt_all(N,Ys,SmallYs),  leq_all(N,Zs,BigZs),
   split(Xs,Ys,Zs).

:- pred ff3.
ff3 :-                              % takedrop
   constr( ~((SmallTs & BigDs) => (BLs = (BTs & BDs))) ),
   is_asorted(Ls,BLs), is_asorted(Ts,BTs), is_asorted(Ds,BDs),
   grt_all(M,Ts,SmallTs),  leq_all(M,Ds,BigDs),
   takedrop(N,Ls,Ts,Ds).
   
:- pred ff4.
ff4 :-                              % merge
   constr( ~((BXs & BYs) => BZs)),
   listmin(Xs,IsDefMinXs,MinXs), listmin(Ys,IsDefMinYs,MinYs),  
   hd(Xs,IsDefHXs,HXs), hd(Ys,IsDefHYs,HYs),                   
   is_asorted(Xs,BXs), is_asorted(Ys,BYs), is_asorted(Zs,BZs),
   merge(Xs,Ys,Zs).

:- pred ff5.                        % cons
ff5 :- constr(~(((~IsDefMinXs => (B3 & MinYs=X)) & 
                (B3 => (MinYs=X))) &
                ((B2 & (X=<MinXs)) => B3))),
  listmin(Xs,IsDefMinXs,MinXs), listmin(Ys,IsDefMinYs,MinYs),
	is_asorted(Xs,B2), is_asorted(Ys,B3), 
  cons(X,Xs,Ys).

% =================================================================
:- query ff1/0.
:- query ff2/0.
:- query ff3/0.
:- query ff4/0.
:- query ff5/0.
% =================================================================

